""" Draw the sequence diagram of a multi-echo gradient echo sequence, with
    bipolar readout gradients.
"""

import copy

import matplotlib.pyplot
import mrsd
import numpy

# Define the tissue and sequence parameters (arbitrary units): T2,
# echo and repetition times, durations of ramp, pulses, encoding and readout,
# length of the echo train
T2 = 10
TE, TR = 10, 40
d_pulse, d_encoding, d_readout, d_ramp = 1, 3, 1, 0.1
train_length = 10

# Create the underlying Matplotlib objects and the diagram
figure, plot = matplotlib.pyplot.subplots(tight_layout=True, figsize=(10,5))
diagram = mrsd.Diagram(plot, ["RF", "$G_z$", "$G_y$", "$G_x$", "Echoes"])

# Add the excitation RF pulse, show the beginning of the next repetition
excitation, excitation_selection = diagram.selective_pulse(
    "RF", "$G_z$", d_pulse, 0.5, 0.5, ramp=d_ramp, center=0)
diagram.annotate("RF", excitation.center+0.2, 1, "90°")
diagram.add("RF", copy.copy(excitation).move(TR))
diagram.add("$G_z$", copy.copy(excitation_selection).move(TR))
diagram.annotate("RF", excitation.center+0.2+TR, 1, "90°")
diagram.interval(0, TR, -2, "TR")

# Add the refocalization RF pulse
refocalization, _ = diagram.selective_pulse(
    "RF", "$G_z$", d_pulse, 1, 0.5, ramp=d_ramp, center=TE/2)
diagram.annotate("RF", refocalization.center+0.2, 1, "180°")
diagram.interval(0, TE/2, -1, "TE/2")

# Add the first echo
echo = diagram.echo("Echoes", d_readout, 1, center=TE)
readout = diagram.gradient("$G_x$", d_readout, 1, d_ramp, center=echo.center)
diagram.interval(0, TE, -1.5, "TE")

# Add the encoding gradients
diagram.add(
    "$G_z$", excitation_selection.adapt(
        d_encoding, -0.5, d_ramp, begin=excitation_selection.end))
diagram.multi_gradient(
    "$G_y$", d_encoding, 0.5, d_ramp, begin=excitation_selection.end)
diagram.add(
    "$G_x$", readout.adapt(
        d_encoding, -0.5, d_ramp, begin=excitation_selection.end))

# Add the other echoes in the train
for echo in range(1, train_length):
    refocalization_selection = diagram.gradient(
        "$G_z$", d_pulse, 0.5, d_ramp, begin=readout.end)
    refocalization = diagram.sinc_pulse(
        "RF", d_pulse, 1, center=refocalization_selection.center)
    diagram.annotate(
        "RF", refocalization.center+0.2, 1, "180°", fontsize=8, color="0.5")
    
    readout = diagram.gradient(
        "$G_x$", d_readout, 1, d_ramp, begin=refocalization_selection.end)
    
    elapsed = readout.center - TE
    echo_amplitude = numpy.exp(-elapsed/T2)
    diagram.echo("Echoes", d_readout, echo_amplitude, center=readout.center)

# Add the T2 decay curve
xs = numpy.linspace(0, readout.end-TE, 100)
ys = numpy.exp(-xs/T2)
plot.plot(xs+TE, ys+diagram.y("Echoes"), color="C0", lw=1)
plot.text(TE+xs[len(xs)//2], 0.5, "$e^{-t/T_2}$", color="C0")

matplotlib.pyplot.show()



import matplotlib.path
import numpy

from .event import Event

def box_envelope(pulse):
    return (
        [pulse.begin, pulse.begin, pulse.end, pulse.end],
        [0, pulse.amplitude, pulse.amplitude, 0])

def gaussian_envelope(pulse):
    support = numpy.linspace(-1, +1, pulse.points)
    xs = support/(support[-1]-support[0])*(pulse.end-pulse.begin)+pulse.center
    
    ys = pulse.amplitude * numpy.exp(-support**2 / (2*pulse.sd**2))
    
    return xs, ys

def sinc_envelope(pulse):
    support = numpy.linspace(-pulse.lobes, +pulse.lobes, pulse.points)
    xs = support/(support[-1]-support[0])*(pulse.end-pulse.begin)+pulse.center
    
    envelope = (
        sinc if pulse.apodization is None
        else globals()[f"{pulse.apodization}_sinc"])
    ys = pulse.amplitude*envelope(pulse.lobes)(support)
    
    return xs, ys

class RFPulse(Event):
    """ RF pulse event, represented by a user-defined envelope
        
        :param duration: duration of the pulse
        :param amplitude: maximum amplitude
        :param envelope: sinc_envelope (default), box_envelope, or
            gaussian_envelope
        :param lobes: number of lobes on each side of a sinc envelope (defaults
            to 3)
        :param apodization: apodization of a sinc envelope (None (default),
            "hann", or "hamming")
        :param sd: standard deviation of a Gaussian envelope (defaults to 0.3)
        :param points: number of points used to draw the envelope (defaults to
            101)
    """
    
    def __init__(self, duration, amplitude, envelope=sinc_envelope, **kwargs):
        self.envelope = envelope
        self.sd = kwargs.pop("sd", 0.3)
        self.lobes = kwargs.pop("lobes", 3)
        self.apodization = kwargs.pop("apodization", None)
        self.points = kwargs.pop("points", 101)
        super().__init__(duration, amplitude, **kwargs)
        
    def get_path(self):
        xs, ys = self.envelope(self)
        return matplotlib.path.Path(numpy.transpose([xs, ys]))
    
    @property
    def _fields(self):
        return {
            x: getattr(self, x) for x in [
                "duration", "amplitude", "begin", "center", "end",
                "envelope", "lobes", "apodization", "points"]}

def apodized_sinc(N, alpha):
    return lambda t: ((1-alpha) + alpha*numpy.cos(numpy.pi*t/N)) * numpy.sinc(t)

sinc = lambda N: apodized_sinc(N, 0)
hann_sinc = lambda N: apodized_sinc(N, 0.5)
hamming_sinc = lambda N: apodized_sinc(N, 25./46.)



import copy

import matplotlib.pyplot
import matplotlib.ticker
import numpy

from .adc import ADC
from .echo import Echo
from .gradient import Gradient
from .multi_gradient import MultiGradient
from . import rf_pulse
from .rf_pulse import RFPulse

class Diagram(object):
    """ MR sequence diagram
        
        :param plot: an instance of matplotlib axes (plot, subplot, etc.)
        :param channels: sequence of channels names in the plot, from top to
            bottom
    """
    
    def __init__(self, plot, channels):
        self._channel_height = 2
        self._channel_gap = 0.2
        
        self.plot = plot
        self.plot.spines[:].set_visible(False)
        
        self._channels = { 
            x: i*(self._channel_height+self._channel_gap)
            for i, x in enumerate(channels[::-1]) }
        
        self.plot.xaxis.set_major_locator(matplotlib.ticker.NullLocator())
        self.plot.xaxis.set_minor_locator(matplotlib.ticker.NullLocator())
        
        self.plot.yaxis.set_major_locator(
            matplotlib.ticker.FixedLocator(sorted(self._channels.values())))
        self.plot.yaxis.set_minor_locator(matplotlib.ticker.NullLocator())
        def channel_formatter(x, pos):
            distances = [numpy.abs(x-value) for value in self._channels.values()]
            return list(self._channels.keys())[numpy.argmin(distances)]
        self.plot.yaxis.set_major_formatter(channel_formatter)
        self.plot.yaxis.set_tick_params(length=0, width=0)
        
        self._background_line_style = {"color": "0.9", "lw": 1, "zorder": -1}
        
        for y in self._channels.values():
            self.plot.axhline(y, **self._background_line_style)
    
    def add(self, channel, event):
        """ Add an event to the specified channel.
        """
        
        event.offset[1] += self._channels[channel]
        self.plot.add_patch(event)
        self.plot.autoscale_view()
    
    def adc(self, channel, *args, **kwargs):
        """ Add an ADC event to the specified channel.
        """
        
        event = ADC(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def echo(self, channel, *args, **kwargs):
        """ Add an echo event to the specified channel.
        """
        
        event = Echo(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def gradient(self, channel, *args, **kwargs):
        """ Add a gradient event to the specified channel.
        """
        
        event = Gradient(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def multi_gradient(self, channel, *args, **kwargs):
        """ Add a multi-gradient event to the specified channel.
        """
        
        event = MultiGradient(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def rf_pulse(self, channel, *args, **kwargs):
        """ Add an RF pulse event to the specified channel.
        """
        
        event = RFPulse(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def gaussian_pulse(self, channel, *args, **kwargs):
        """ Add a hard RF pulse event to the specified channel.
        """
        
        kwargs["envelope"] = rf_pulse.gaussian_envelope
        event = RFPulse(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def hard_pulse(self, channel, *args, **kwargs):
        """ Add a hard RF pulse event to the specified channel.
        """
        
        kwargs["envelope"] = rf_pulse.box_envelope
        event = RFPulse(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def sinc_pulse(self, channel, *args, **kwargs):
        """ Add a sinc RF pulse event to the specified channel.
        """
        
        kwargs["envelope"] = rf_pulse.sinc_envelope
        event = RFPulse(*args, **kwargs)
        self.add(channel, event)
        return event
    
    def readout(
            self, adc_channel, gradient_channel, duration,
            echo_amplitude=1, gradient_amplitude=1,
            ramp=0, ramp_up=None, ramp_down=None, 
            begin=None, end=None, center=None,
            adc_kwargs=None, echo_kwargs=None, gradient_kwargs=None):
        """ Add a readout block (echo, ADC and gradient)
            
            :param adc_channel: channel of the echo and ADC
            :param gradient_channel: channel of the gradient
            :param duration: duration of the echo, ADC, and gradient flat-top
            :param echo_amplitude: amplitude of the echo
            :param gradient_amplitude: amplitude of the gradient flat-top
            :param ramp,ramp_up,ramp_down: ramp durations of the gradient.
                Use `ramp` for symmetric gradients, and both `ramp_up` and
                `ramp_down` for asymmetric gradients
            :param begin,end,center: time of the begin, end, or center of the
                echo/ADC. Only one must be specified.
            :param adc_kwargs: extra parameters for the ADC event (e.g. style)
            :param echo_kwargs: extra parameters for the echo event (e.g. style)
            :param gradient_kwargs: extra parameters for the gradient event
                (e.g. style)
        """
        
        adc = self.adc(
            adc_channel, duration, begin=begin, end=end, center=center,
            **(adc_kwargs or {}))
        echo = self.echo(
            adc_channel, adc.duration, echo_amplitude, center=adc.center,
            **(echo_kwargs or {}))
        gradient = self.gradient(
            gradient_channel, adc.duration, gradient_amplitude,
            ramp_up=ramp_up if ramp_up is not None else ramp,
            ramp_down=ramp_down if ramp_down is not None else ramp,
            center=adc.center, **(gradient_kwargs or {}))
        
        return adc, echo, gradient
    
    def selective_pulse(
            self, pulse_channel, gradient_channel, duration,
            pulse_amplitude=1, gradient_amplitude=1,
            envelope=None, ramp=0, ramp_up=None, ramp_down=None, 
            begin=None, end=None, center=None,
            pulse_kwargs=None, gradient_kwargs=None):
        """ Add a selective pulse block (pulse and gradient)
            
            :param pulse_channel: channel of the RF pulse
            :param gradient_channel: channel of the gradient
            :param duration: duration of the pulse and gradient flat-top
            :param pulse_amplitude: amplitude of the RF pulse
            :param gradient_amplitude: amplitude of the gradient flat-top
            :param envelope: envelope of the pulse (default to sinc)
            :param ramp,ramp_up,ramp_down: ramp durations of the gradient.
                Use `ramp` for symmetric gradients, and both `ramp_up` and
                `ramp_down` for asymmetric gradients
            :param begin,end,center: time of the begin, end, or center of the
                echo/ADC. Only one must be specified.
            :param pulse_kwargs: extra parameters for the pulse event (e.g.
                style)
            :param gradient_kwargs: extra parameters for the gradient event
                (e.g. style)
        """
        
        pulse = self.rf_pulse(
            pulse_channel, duration, pulse_amplitude,
            envelope or rf_pulse.sinc_envelope,
            begin=begin, end=end, center=center,
            **(pulse_kwargs or {}))
        gradient = self.gradient(
            gradient_channel, pulse.duration, gradient_amplitude,
            ramp_up=ramp_up if ramp_up is not None else ramp,
            ramp_down=ramp_down if ramp_down is not None else ramp,
            center=pulse.center, **(gradient_kwargs or {}))
        
        return pulse, gradient
    
    def annotate(self, channel, x, y, text, **kwargs):
        """ Add an annotation
            
            :param channel: channel to which the annotation is added
            :param x: time of the annotation
            :param y: relative position of the annotation in the channel
            :param text: text of the annotation
            :param kwargs: extra parameters passed to matplotlib.axes.Axes.text
        """
        
        self.plot.text(x, self._channels[channel]+y, text, **kwargs)
    
    def interval(self, begin, end, y, label, color="k"):
        """ Add a time interval annotation
            
            :param begin,end: begin and end time of the interval
            :param y: vertical position of the annotation
            :param label: label of the annotation
            :param color: color of the annotation label
        """
        
        self.plot.set(ylim=min(y, self.plot.get_ylim()[0]))
        
        self.plot.annotate(
            "", (begin, y), (end, y),
            arrowprops={
                "arrowstyle":"<|-|>", "shrinkA": 0, "shrinkB": 0,
                **self._background_line_style})
        self.plot.text(
            (begin+end)/2, y, label,
            color=color, bbox={"fc": "white", "ec": "none"},
            horizontalalignment="center", verticalalignment="center")
        
        self.plot.vlines(
            [begin, end],
            y, self._channel_height/2+max(self._channels.values()),
            **self._background_line_style)
    
    def y(self, channel):
        """ Return the y coordinate of the center of a channnel.
        """
        
        return self._channels[channel]



""" Draw the sequence diagram of the FLASH sequence. Adapted from "FLASH
    imaging. Rapid NMR imaging using low flip-angle pulses", Haase et al.,
    Journal of Magnetic Resonance, 67(2), pp. 258-266, 1986.
"""

import copy

import matplotlib.pyplot
import mrsd

# Define sequence parameters (arbitrary units): echo and repetition times,
# durations of ramp, pulses, encoding and readout
TE, TR = 4, 10
d_ramp, d_pulse, d_encoding, d_readout = 0.1, 1, 1, 2

# Create the underlying Matplotlib objects and the diagram
figure, plot = matplotlib.pyplot.subplots(tight_layout=True)
diagram = mrsd.Diagram(
    plot, ["RF", "$G_{slice}$", "$G_{phase}$", "$G_{readout}$", "Signal"])

# Slice-selective pulse of the first TR
excitation, slice_selection = diagram.selective_pulse(
    "RF", "$G_{slice}$", d_pulse, ramp=d_ramp, center=0)

# Readout, centered on TE
adc, echo, readout = diagram.readout(
    "Signal", "$G_{readout}$", d_readout, ramp=d_ramp, center=TE,
    adc_kwargs={"ec": "0.5"})

# Encoding: rewind slice-selection gradient, run phase gradient, prephase
# read-out gradient
diagram.add(
    "$G_{slice}$",
    slice_selection.adapt(d_encoding, -0.5, d_ramp, end=readout.begin))
diagram.multi_gradient(
    "$G_{phase}$", d_encoding, 1, d_ramp, end=readout.begin)
diagram.add(
    "$G_{readout}$", readout.adapt(d_encoding, -0.5, d_ramp, end=readout.begin))

# Start of next TR
diagram.add("RF", copy.copy(excitation).move(TR))
diagram.add("$G_{slice}$", copy.copy(slice_selection).move(TR))

# Add annotations: flip angles and TE/TR intervals
diagram.annotate("RF", 0.2, 1, r"$\alpha$")
diagram.annotate("RF", TR+0.2, 1, r"$\alpha$")
diagram.interval(0, TE, -1.5, "TE")
diagram.interval(0, TR, -2.5, "TR")

matplotlib.pyplot.show()



from .event import Event
from .adc import ADC
from .echo import Echo
from .gradient import Gradient
from .multi_gradient import MultiGradient
from .rf_pulse import RFPulse, box_envelope, gaussian_envelope, sinc_envelope

from .diagram import Diagram



import matplotlib.path
import numpy

from .event import Event

class MultiGradient(Event):
    """ Multiple gradient events (e.g. phase encoding), represented by nested
        trapezoids.
        
        :param flat_top: duration of the gradient flat-top
        :param amplitude: maximum amplitude of the gradient flat-top
        :param ramp,ramp_up,ramp_down: ramp durations of the gradient.
            Use `ramp` for symmetric gradients, and both `ramp_up` and
            `ramp_down` for asymmetric gradients
        :param steps: number of steps drawn between the maximum and minimum
            amplitude
    """
    
    def __init__(self, flat_top, amplitude, ramp=0, steps=5, **kwargs):
        
        kwargs.setdefault("ramp_up", ramp)
        kwargs.setdefault("ramp_down", ramp)
        
        self.flat_top = flat_top
        self.ramp_up = kwargs.pop("ramp_up")
        self.ramp_down = kwargs.pop("ramp_down")
        
        self.steps = steps
        
        super().__init__(
            self.ramp_up+self.flat_top+self.ramp_down, amplitude, **kwargs)
        
    def get_path(self):
        amplitudes = numpy.linspace(-self.amplitude, self.amplitude, self.steps)
        
        paths = [
            matplotlib.path.Path([
                [self.begin, 0],
                [self.begin+self.ramp_up, amplitude],
                [self.begin+self.ramp_up+self.flat_top, amplitude],
                [self.end, 0]])
            for amplitude in [amplitudes[0], amplitudes[-1]]]
        
        # θ: angle at the top of the triangle forming the ramp of the gradient
        # sin θ = ramp / amplitude ⇒ θ = arcsin(ramp/amplitude)
        # at height h, the side is from h to amplitude-h ⇒ the base is
        # sin θ (amplitude - h)
        theta_up = numpy.arcsin(self.ramp_up / self.amplitude)
        theta_down = numpy.arcsin(self.ramp_down / self.amplitude)
        
        for amplitude in amplitudes[1:-1]:
            extra_up = numpy.sin(theta_up) * numpy.abs(amplitude)
            extra_down = numpy.sin(theta_down) * numpy.abs(amplitude)
            paths.append(matplotlib.path.Path([
                [self.begin+extra_up, amplitude],
                [self.end-extra_down, amplitude]]))
        
        return matplotlib.path.Path.make_compound_path(*paths)
    
    @property
    def _fields(self):
        return {
            x: getattr(self, x)
            for x in [
                "flat_top", "amplitude", "ramp_up", "ramp_down", "steps",
                "begin"]}



import matplotlib.path

from .event import Event

class Gradient(Event):
    """ Gradient event, represented by a trapezoid.
        
        :param flat_top: duration of the gradient flat-top
        :param amplitude: amplitude of the gradient flat-top
        :param ramp,ramp_up,ramp_down: ramp durations of the gradient.
            Use `ramp` for symmetric gradients, and both `ramp_up` and
            `ramp_down` for asymmetric gradients
    """
    
    def __init__(self, flat_top, amplitude, ramp=0, **kwargs):
        kwargs.setdefault("ramp_up", ramp)
        kwargs.setdefault("ramp_down", ramp)
        
        self.flat_top = flat_top
        self.ramp_up = kwargs.pop("ramp_up")
        self.ramp_down = kwargs.pop("ramp_down")
        
        super().__init__(
            self.ramp_up+self.flat_top+self.ramp_down, amplitude, **kwargs)
    
    def adapt(
            self, flat_top, area_factor,
            ramp=0, ramp_up=None, ramp_down=None, **kwargs):
        """ Return a gradient with an area equal to a factor of the current
            gradient.
        """
        
        area = self.amplitude * (self.ramp_up/2 + self.flat_top + self.ramp_down/2)
        target_area = area * area_factor
        
        if ramp_up is None:
            ramp_up = ramp
        if ramp_down is None:
            ramp_down = ramp
        amplitude = target_area / (ramp_up/2 + flat_top + ramp_down/2)
        
        return Gradient(
            flat_top, amplitude, ramp_up=ramp_up, ramp_down=ramp_down, **kwargs)
    
    def get_path(self):
        return matplotlib.path.Path([
            [self.begin, 0],
            [self.begin+self.ramp_up, self.amplitude],
            [self.begin+self.ramp_up+self.flat_top, self.amplitude],
            [self.end, 0]
        ])
    
    @property
    def _fields(self):
        return {
            x: getattr(self, x)
            for x in ["flat_top", "amplitude", "ramp_up", "ramp_down", "begin"]}



import os

import setuptools

here = os.path.abspath(os.path.dirname(__file__))
long_description = open(os.path.join(here, "README.md")).read()

setuptools.setup(
    name="mrsd",
    version="1.0.0",
    
    description="MR sequence diagram toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    
    url="https://github.com/lamyj/mrsd",
    
    author="Julien Lamy",
    author_email="lamy@unistra.fr",
    
    license="MIT",
    
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Framework :: Matplotlib",
        "Intended Audience :: Education",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    
    keywords = ["MRI", "sequence", "diagram", "matplotlib"],
    
    packages=["mrsd"],
    package_dir={"mrsd": "src/mrsd"},
    
    python_requires=">=3.6",
    
    install_requires=["matplotlib>=3.3", "numpy"],
)



import matplotlib.path
import numpy

from .event import Event

class Echo(Event):
    """ Echo event, represented by an oscillating exponential.
    """
    
    def __init__(self, duration, amplitude=1, **kwargs):
        super().__init__(duration, amplitude, **kwargs)
    
    def get_path(self):
        npoints = 26
        slope = numpy.linspace(0, 1, npoints)
        sign = -1+2*(numpy.arange(npoints)%2)
        
        ys = sign * numpy.exp(slope*4)
        # Make symmetrical
        ys = numpy.concatenate((ys, ys[-2::-1]))
        # Normalize amplitude and taper the ends
        ys *= self.amplitude/numpy.abs(ys).max()
        ys[0] = ys[-1] = 0
        
        xs = numpy.linspace(self.begin, self.end, 2*npoints-1)
        
        return matplotlib.path.Path(numpy.transpose([xs, ys]))
    
    @property
    def _fields(self):
        return {x: getattr(self, x) for x in ["duration", "amplitude", "begin"]}



import matplotlib.patches
import matplotlib.transforms
import numpy

class Event(matplotlib.patches.Patch):
    """ Abstract sequence event
        
        :param duration: total duration.
        :param amplitude: normalized amplitude between -1 and +1.
        :param begin,end,center: time of the begin, end, or center of the
            event. Only one must be specified.
        :param offset: horizontal and vertical offset to position the event.
        :param kwargs: passed to matplotlib.patches.Patch
    """
    
    def __init__(
            self, duration, amplitude,
            begin=None, end=None, center=None, offset=None,
            **kwargs):
        self.duration = duration
        self.amplitude = amplitude
        
        if begin is not None:
            self.begin = begin
        elif center is not None:
            self.begin = center - self.duration/2
        elif end is not None:
            self.begin = end - self.duration
        else:
            raise Exception("Missing time")
        
        self.center = self.begin + self.duration/2
        self.end = self.begin + self.duration
        
        self.offset = numpy.array(offset if offset is not None else [0., 0.])
        
        if "color" not in kwargs:
            if"edgecolor" not in kwargs and "ec" not in kwargs:
                kwargs.setdefault("edgecolor", "black")
            if "facecolor" not in kwargs and "fc" not in kwargs:
                kwargs.setdefault("facecolor", "none")
        if "linewidth" not in kwargs and "lw" not in kwargs:
            kwargs.setdefault("linewidth", 1)
        
        super().__init__(**kwargs)
    
    def move(self, offset):
        """ Move on the time axis, return the object
        """
        
        self.begin += offset
        self.center += offset
        self.end += offset
        return self
    
    def __copy__(self):
        new_object = self.__class__(**self._fields)
        inspector = matplotlib.artist.ArtistInspector(self)
        setters = inspector.get_setters()
        for name, value in inspector.properties().items():
            if name == "transform" or name not in setters:
                continue
            
            getattr(new_object, f"set_{name}")(value)
        
        return new_object
    
    def get_patch_transform(self):
        """ Inherited from parent class.
        """
        
        return matplotlib.transforms.Affine2D().translate(*self.offset)



import matplotlib.path

from .event import Event

class ADC(Event):
    """ ADC/readout event, represented by a rectangle of amplitude 1.
    """
    
    def __init__(self, duration, **kwargs):
        super().__init__(duration, 1, **kwargs)
    
    def get_path(self):
        return matplotlib.path.Path([
            [self.begin, 0],
            [self.begin, self.amplitude],
            [self.begin+self.duration, self.amplitude],
            [self.begin+self.duration, 0]])
    
    @property
    def _fields(self):
        return {x: getattr(self, x) for x in ["duration", "begin"]}
